package cs155.hw3;


/**
   This is now an abstract class since all objects have inside and outside materials
**/

public class TransformedObject3D extends Object3D {

    Object3D obj;
    Transform3D trans = new Transform3D();
    Transform3D invtrans = new Transform3D();

    /**
       Generate an object which can be transformed by any combination of translation, scaling, and rotations.
     **/
    public TransformedObject3D(Object3D obj) {
	this.obj=obj;
        this.insideMat = obj.insideMat;
        this.outsideMat = obj.outsideMat;
    }

    private TransformedObject3D(Object3D obj, Transform3D trans, Transform3D invtrans) {
	this.obj=obj; this.trans=trans; this.invtrans=invtrans;
        this.insideMat = obj.insideMat;
        this.outsideMat = obj.outsideMat;

    }

    //  x.rayIntersect(ray) -- returns a double, -1 for no intersection, t>0 for intersection at value t
    /**
       Intersect a ray with a transformed object.
       To do this, we intersect the object with the inverse transformed ray
       and get a distance and normal for the intersection. We then need to adjust
       the distance and apply the transpose of the inverse transform to the normal!!
       to get the RayHit for the transformed object.
     **/
    public RayHit rayIntersect (Ray3D ray) {
        Ray3D newRay = invtrans.apply(ray);
        RayHit hit = obj.rayIntersect(newRay);
        if (hit.distance == Double.POSITIVE_INFINITY) return hit;

	double d1 = invtrans.apply(ray.d).length();
        double t0 = hit.distance/d1;
        Point3D n0 = invtrans.transpose().apply(hit.normal); // tricky!!
        return new RayHit(t0,n0,hit.obj);
    }

    /**
       Translate an object by the vector defined by a point p.
     **/
    public TransformedObject3D translate(Point3D p) {
        Transform3D 
	    op = Transform3D.translateOp(p.x, p.y, p.z),
	    invOp = Transform3D.translateOp(-p.x, -p.y, -p.z);
	return new TransformedObject3D(obj, op.mult(trans), invtrans.mult(invOp));
    }

    /**
       Scale the object along its x,y, and z axes.
     **/
    public TransformedObject3D scale(double x, double y, double z) {
        Transform3D
	    op = Transform3D.scaleOp(x,y,z),
	    invOp = Transform3D.scaleOp(1/x,1/y,1/z);

	return new TransformedObject3D(obj, op.mult(trans), invtrans.mult(invOp));
    }

    /**
       Rotate the object around its x axis.
     **/
    public TransformedObject3D rotateX(double angle){
        Transform3D
	    op = Transform3D.rotateXOp(angle),
	    invOp = Transform3D.rotateXOp(-angle);

	return new TransformedObject3D(obj, op.mult(trans), invtrans.mult(invOp));
    }

    /**
       Rotate the object around its y axis.
     **/
    public TransformedObject3D rotateY(double angle){
        Transform3D
	    op = Transform3D.rotateYOp(angle),
	    invOp = Transform3D.rotateYOp(-angle);

	return new TransformedObject3D(obj, op.mult(trans), invtrans.mult(invOp));

    }

    /**
       Rotate the object around its z axis.
     **/
    public TransformedObject3D rotateZ(double angle){
        Transform3D
	    op = Transform3D.rotateZOp(angle),
	    invOp = Transform3D.rotateZOp(-angle);

	return new TransformedObject3D(obj, op.mult(trans), invtrans.mult(invOp));
    }



}


